#!/usr/bin/env python

##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#                    Arnaud Fontaine <arnaud.fontaine@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import argparse
import os
import sys
import multiprocessing
import errno

from .argument import ArgumentType
from .process import BenchmarkProcess
from .result import ERP5BenchmarkResult, CSVBenchmarkResult

MAXIMUM_KEYBOARD_INTERRUPT = 5

class PerformanceTester(object):
  def __init__(self, namespace=None):
    if not namespace:
      self._argument_namespace = self._parse_arguments(argparse.ArgumentParser(
          description='Run ERP5 benchmarking suites.'))
    else:
      self._argument_namespace = namespace

  @staticmethod
  def _add_parser_arguments(parser):
    # Optional arguments
    parser.add_argument('--filename-prefix',
                        default='result',
                        metavar='PREFIX',
                        help='Filename prefix for results and logs files '
                             '(default: result)')

    parser.add_argument('--report-directory', '-r',
                        type=ArgumentType.directoryType,
                        default=os.getcwd(),
                        metavar='DIRECTORY',
                        help='Directory where the results and logs will be stored '
                             '(default: current directory)')

    parser.add_argument('--max-global-average',
                        type=float,
                        default=0,
                        metavar='N',
                        help='Stop when any suite operation is over this value '
                             '(default: disable)')

    parser.add_argument('--users-file',
                        dest='user_info_filename',
                        default='userInfo',
                        metavar='MODULE',
                        help="Import users from ``user_tuple'' in MODULE")

    parser.add_argument('--users-range-increment',
                        type=ArgumentType.checkIntValueWrapper(minimum=1),
                        default=1,
                        metavar='N',
                        help='Number of users being added after each repetition '
                             '(default: 1)')

    parser.add_argument('--max-errors',
                        dest='max_error_number',
                        type=ArgumentType.checkIntValueWrapper(minimum=0),
                        default=10,
                        help='Stop execution after N consecutive errors '
                             '(default: 10)')

    parser.add_argument('--enable-debug', '-d',
                        action='store_true',
                        default=False,
                        help='Enable debug messages')

    parser.add_argument('--enable-legacy-listbox',
                        dest='is_legacy_listbox',
                        action='store_true',
                        default=False,
                        help='Enable legacy listbox for Browser')

    parser.add_argument('--repeat',
                        type=ArgumentType.checkIntValueWrapper(minimum=1),
                        default=-1,
                        metavar='N',
                        help='Repeat the benchmark suite N times for a given '
                             'number of users (default: infinite)')

    parser.add_argument('--repeat-range',
                        type=ArgumentType.checkIntValueWrapper(minimum=1),
                        default=-1,
                        metavar='N',
                        help='Repeat the benchmark suite N times for the whole '
                             'range of users (default: infinite)')

    parser.add_argument('--user-index',
                        type=int,
                        default=0,
                        metavar='INDEX',
                        help='Index of the first user within userInfo '
                             '(default: 0)')

    parser.add_argument('--erp5-publish-url',
                        metavar='ERP5_PUBLISH_URL',
                        help='ERP5 URL to publish the results to '
                             '(default: disabled, thus writing to CSV files)')

    parser.add_argument('--erp5-publish-project',
                        metavar='ERP5_PUBLISH_PROJECT',
                        help='ERP5 publish project')

    # Mandatory arguments
    parser.add_argument('erp5_base_url', metavar='ERP5_URL')

    parser.add_argument('users',
                        type=ArgumentType.strictlyPositiveIntOrRangeType,
                        metavar='NB_USERS|MIN_NB_USERS,MAX_NB_USERS',
                        help='Number of users (fixed or a range)')

    parser.add_argument('benchmark_suite_list',
                        nargs='+',
                        metavar='BENCHMARK_SUITES',
                        help='Benchmark suite modules')

  @staticmethod
  def _check_parsed_arguments(namespace):
    namespace.user_tuple = ArgumentType.objectFromModule(namespace.user_info_filename,
                                                         object_name='user_tuple')

    object_benchmark_suite_list = []
    for benchmark_suite in namespace.benchmark_suite_list:
      object_benchmark_suite_list.append(ArgumentType.objectFromModule(benchmark_suite,
                                                                       callable_object=True))

    if namespace.repeat > 0:
      namespace.max_error_number = \
          min(len(namespace.benchmark_suite_list) * namespace.repeat,
              namespace.max_error_number)

    if isinstance(namespace.users, tuple) and namespace.repeat == -1:
        raise argparse.ArgumentTypeError("Repeat cannot be infinite for a "
                                         "range of users")

    namespace.benchmark_suite_name_list = namespace.benchmark_suite_list
    namespace.benchmark_suite_list = object_benchmark_suite_list

    max_nb_users = isinstance(namespace.users, tuple) and namespace.users[1] or \
        namespace.users

    namespace.user_tuple = namespace.user_tuple[namespace.user_index:]
    if max_nb_users > len(namespace.user_tuple):
      raise argparse.ArgumentTypeError("Not enough users in the given file")

    if (namespace.erp5_publish_url and not namespace.erp5_publish_project) or \
       (not namespace.erp5_publish_url and namespace.erp5_publish_project):
      raise argparse.ArgumentTypeError("Publish ERP5 URL and project must "
                                       "be specified")

    return namespace

  @staticmethod
  def _parse_arguments(parser):
    PerformanceTester._add_parser_arguments(parser)
    namespace = parser.parse_args()
    PerformanceTester._check_parsed_arguments(namespace)
    return namespace

  def getResultClass(self):
    if self._argument_namespace.erp5_publish_url:
      return ERP5BenchmarkResult
    else:
      return CSVBenchmarkResult

  def preRun(self):
    if not self._argument_namespace.erp5_publish_url:
      return

    self._argument_namespace.erp5_publish_url += \
        ERP5BenchmarkResult.createResultDocument(self._argument_namespace.erp5_publish_url,
                                                 self._argument_namespace.erp5_publish_project,
                                                 self._argument_namespace.repeat,
                                                 self._argument_namespace.users)          

  def postRun(self, error_message_set):
    if not self._argument_namespace.erp5_publish_url:
      return

    ERP5BenchmarkResult.closeResultDocument(self._argument_namespace.erp5_publish_url,
                                            error_message_set)

  def _run_constant(self, nb_users, current_repeat_range=None):
    process_list = []
    exit_msg_queue = multiprocessing.Queue(nb_users)

    result_class = self.getResultClass()

    for user_index in range(nb_users):
      process = BenchmarkProcess(exit_msg_queue, result_class,
                                 self._argument_namespace, nb_users,
                                 user_index,
                                 current_repeat_range)

      process_list.append(process)

    for process in process_list:
      process.start()

    error_message_set = set()
    process_terminated_counter = 0

    # Ensure that SIGTERM signal (sent by terminate()) is not sent twice
    exit_status = 0

    while process_terminated_counter != len(process_list):
      try:
        error_message = exit_msg_queue.get()

      except KeyboardInterrupt, e:
        print >>sys.stderr, "\nInterrupted by user, stopping gracefully..."
        exit_status = 2

      # An IOError may be raised  when receiving a SIGINT which interrupts the
      # blocking system call above and the system call should not be restarted
      # (using siginterrupt), otherwise the  process will stall forever as its
      # child has already exited
      except IOError, e:
        if e.errno == errno.EINTR:
          continue

      else:
        if error_message is not None:
          error_message_set.add(error_message)
          exit_status = 1

        process_terminated_counter += 1

      # In case of  error or SIGINT, kill the other  children because they are
      # likely failing as well (especially  because a process only exits after
      # encountering 10 errors)
      if exit_status != 0:
        for process in process_list:
          if process.is_alive():
            process.terminate()
            process.join()

    return (error_message_set, exit_status)

  def run(self):
    error_message_set, exit_status = set(), 0
    self.preRun()

    if isinstance(self._argument_namespace.users, tuple):
      min_user_number, max_user_number = self._argument_namespace.users
      repeat_counter = 0
      error_counter = 0

      while (repeat_counter != self._argument_namespace.repeat_range and
             error_counter != self._argument_namespace.max_error_number and
             exit_status != 2):
        current_user_number = min_user_number

        while True:
          current_error_message_set, exit_status = \
              self._run_constant(current_user_number, repeat_counter)

          if exit_status != 0:
            error_counter += 1
            error_message_set.update(current_error_message_set)

          if (current_user_number == max_user_number or
              error_counter == self._argument_namespace.max_error_number or
              exit_status == 2):
            break

          current_user_number = \
              min((current_user_number +
                   self._argument_namespace.users_range_increment),
                  max_user_number)

        repeat_counter += 1

    else:
      error_message_set, exit_status = self._run_constant(
        self._argument_namespace.users)

    self.postRun(error_message_set)
    return error_message_set, exit_status

def main():
  error_message_set, exit_status = PerformanceTester().run()
  for error_message in error_message_set:
    print >>sys.stderr, "ERROR: %s" % error_message

  sys.exit(exit_status)

if __name__ == '__main__':
  main()



#!/usr/bin/env python

##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#                    Arnaud Fontaine <arnaud.fontaine@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from .result import CSVBenchmarkResult, NothingFlushedException

class CSVScalabilityBenchmarkResult(CSVBenchmarkResult):
  def flush(self, partial=True):
    try:
      super(CSVScalabilityBenchmarkResult, self).flush(partial)
    except NothingFlushedException:
      pass
    else:
      self._argument_namespace.notify_method(self._result_filename,
                                             self._result_file.tell(),
                                             partial=partial)

from .performance_tester import PerformanceTester

class ScalabilityTester(PerformanceTester):
  def preRun(self, *args, **kwargs):
    pass

  def postRun(self, error_message_set):
    from logging import Formatter
    import sys
    import urllib
    import urllib2

    try:
      urllib2.urlopen("http://[%s]:%d/report" % \
                        (self._argument_namespace.manager_address,
                         self._argument_namespace.manager_port),
                      urllib.urlencode({'error_message_set': '|'.join(error_message_set)})).close()

    except:
      print >>sys.stderr, "ERROR: %s" % Formatter().formatException(sys.exc_info())

  def getResultClass(self):
    if not self._argument_namespace.erp5_publish_url:
      return CSVScalabilityBenchmarkResult

    return super(ScalabilityTester, self).getResultClass()

from slapos.tool.nosqltester import NoSQLTester

class RunScalabilityTester(NoSQLTester):
  def __init__(self):
    super(RunScalabilityTester, self).__init__()

  def _add_parser_arguments(self, parser):
    super(RunScalabilityTester, self)._add_parser_arguments(parser)
    ScalabilityTester._add_parser_arguments(parser)

  def _parse_arguments(self, parser):
    namespace = super(RunScalabilityTester, self)._parse_arguments(parser)
    ScalabilityTester._check_parsed_arguments(namespace)
    namespace.notify_method = self.send_result_availability_notification
    return namespace

  def run_tester(self):
    ScalabilityTester(self.argument_namespace).run()

def main():
  RunScalabilityTester().run()

if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#                    Arnaud Fontaine <arnaud.fontaine@nexedi.com>
#
# First version: ERP5Mechanize from Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import logging
import sys
import urllib

from urlparse import urljoin
from z3c.etestbrowser.browser import ExtendedTestBrowser
from zope.testbrowser.browser import onlyOne

def measurementMetaClass(prefix):
  """
  Prepare a meta class where the C{prefix} is used to select for which
  methods measurement methods will be added automatically.

  @param prefix:
  @type prefix: str
  @return: The measurement meta class corresponding to the prefix
  @rtype: type
  """
  class MeasurementMetaClass(type):
    """
    Meta class to automatically wrap methods whose prefix starts with
    C{prefix}, and also to define C{lastRequestSeconds} on other classes
    besides of Browser.
    """
    def __new__(metacls, name, bases, dictionary):
      def nonBrowserClass_getattr(self, attribute):
        """
        Some attributes, such as lastRequestSeconds, are only defined in the
        Browser class
        """
        return getattr(self.browser, attribute)

      def timeInSecondDecorator(method):
        def wrapper(self, *args, **kwargs):
          """
          Replaced by the wrapped method docstring. Some methods return the
          time spent on waiting (C{submitSelectJump} and for example), thus
          return a tuple with the time spent on the request and the time spent
          on waiting
          """
          if 'sleep' in kwargs:
            minimum_sleep, maximum_sleep = kwargs.pop('sleep')
            self.randomSleep(minimum_sleep, maximum_sleep)

          ret = method(self, *args, **kwargs)
          if ret is None:
            return self.lastRequestSeconds
          elif isinstance(ret, tuple):
            return (self.lastRequestSeconds,) + ret
          else:
            return (self.lastRequestSeconds, ret)

        return wrapper

      def applyMeasure(method):
        """
        Inner function to wrap timed methods to automatically return the time
        spent on the request

        @param method: Instance method to be called
        @type method: function
        """
        wrapper_method = timeInSecondDecorator(method)
        wrapper_method.func_name = method.func_name
        wrapper_method.__doc__ = method.__doc__

        # In order to avoid re-wrapping the method when looking at the bases
        # for example
        wrapper_method.__is_wrapper__ = True

        dictionary[method.func_name] = wrapper_method

      # Only wrap methods prefixed by the given prefix
      for attribute_name, attribute in dictionary.items():
        if attribute_name.startswith(prefix) and callable(attribute):
          applyMeasure(attribute)

      # And also create these methods by looking at the bases
      for attribute_name in dir(bases[0]):
        if attribute_name not in dictionary and \
           attribute_name.startswith(prefix):
          attribute = getattr(bases[0], attribute_name)
          if callable(attribute) and not getattr(attribute, '__is_wrapper__', False):
            applyMeasure(attribute)

      if 'Browser' not in bases[0].__name__:
        dictionary['__getattr__'] = nonBrowserClass_getattr

      return super(MeasurementMetaClass,
                   metacls).__new__(metacls, name, bases, dictionary)

  return MeasurementMetaClass

import random
import time

class Browser(ExtendedTestBrowser):
  """
  Implements mechanize tests specific to an ERP5 environment through
  U{ExtendedTestBrowser<http://pypi.python.org/pypi/z3c.etestbrowser>}
  (providing features to parse XML and access elements using XPATH)
  using U{zope.testbrowser<http://pypi.python.org/pypi/zope.testbrowser>}
  (providing benchmark and testing features on top of
  U{mechanize<http://wwwsearch.sourceforge.net/mechanize/>}).

  @todo:
   - getFormulatorFieldValue
  """
  __metaclass__ = measurementMetaClass(prefix='open')

  def __init__(self,
               erp5_base_url,
               username,
               password,
               log_file=None,
               is_debug=False,
               is_legacy_listbox=False):
    """
    Create a browser object.

    @param erp5_base_url: ERP5 HTTP URL
    @type erp5_base_url: str
    @param username: Username to be used to log into ERP5
    @type username: str
    @param password: Password to be used to log into ERP5
    @param log_file: Log file object (stderr if none given)
    @type log_file: file
    @param is_debug: Enable or disable debugging (disable by default)
    @type is_debug: bool
    @param is_legacy_listbox: Use legacy listbox
    @type is_legacy_listbox: bool
    """
    # Meaningful to re-create the MainForm class every time the page
    # has been changed
    self._main_form = None

    self._erp5_base_url = erp5_base_url
    if self._erp5_base_url[-1] != '/':
      self._erp5_base_url += '/'

    self._username = username
    self._password = password

    # Only display WARNING message if debugging is not enabled
    logging_level = is_debug and logging.DEBUG or logging.WARNING
    stream = log_file and log_file or sys.stderr
    logging.basicConfig(
      stream=stream, level=logging_level,
      format='%(asctime)s: %(levelname)s: %(message)s [%(name)s]')

    self._logger = logging.getLogger('erp5.util.testbrowser')
    self._is_legacy_listbox = is_legacy_listbox
    self._is_logged_in = False

    super(Browser, self).__init__()

  def open(self, url_or_path=None, data=None, site_relative=True):
    """
    Open a relative (to the ERP5 base URL) or absolute URL. If the given URL
    is not given, then it will open the home ERP5 page. If C{site_relative} is
    False, it will open the URL within the current context.

    @param url_or_path: Relative or absolute URL
    @type url_or_path: str
    """
    if site_relative:
      # In case url_or_path is an absolute URL, urljoin() will return
      # it, otherwise it is a relative path and will be concatenated to
      # ERP5 base URL
      url_or_path = urljoin(self._erp5_base_url, url_or_path)

    if isinstance(data, dict):
      data = urllib.urlencode(data)

    self._logger.debug("Opening: " + url_or_path)
    super(Browser, self).open(url_or_path, data)

  def openNoVisit(self, url_or_path, data=None, site_relative=True):
    """
    Copy/paste from zope.testbrowser.Browser.open() to allow opening an URL
    without changing the current page. See L{open}.

    @see zope.testbrowser.interfaces.IBrowser
    """
    if site_relative:
      # In case url_or_path is an absolute URL, urljoin() will return
      # it, otherwise it is a relative path and will be concatenated to
      # ERP5 base URL
      url_or_path = urljoin(self._erp5_base_url, url_or_path)

    import mechanize

    if isinstance(data, dict):
      data = urllib.urlencode(data)

    response = None
    url_or_path = str(url_or_path)
    self._logger.debug("Opening: " + url_or_path)
    self._start_timer()
    try:
      try:
        try:
          response = self.mech_browser.open_novisit(url_or_path, data)
        except Exception, e:
          fix_exception_name(e)
          raise
      except mechanize.HTTPError, e:
        if e.code >= 200 and e.code <= 299:
          # 200s aren't really errors
          pass
        elif self.raiseHttpErrors:
          raise
    finally:
      self._stop_timer()

    # if the headers don't have a status, I suppose there can't be an error
    if 'Status' in self.headers:
      code, msg = self.headers['Status'].split(' ', 1)
      code = int(code)
      if self.raiseHttpErrors and code >= 400:
        raise mechanize.HTTPError(url_or_path, code, msg, self.headers, fp=None)

    return response

  def randomSleep(self, minimum, maximum):
    """
    Allow to randomly sleep in seconds in the interval [minimum,
    maximum] and is meaningful to simulate an user more realistically.

    It is also called automatically when passing sleep argument to
    click* (C{Link} classes), submit* (C{Form} classes) or open*
    (C{Browser} classes) functions, see C{timeInSecondDecorator} wrapper.

    @param minimum: Minimum number of seconds to sleep
    @type minimum: int
    @param maximum: Maximum number of seconds to sleep
    @type maximum: int
    """
    sleep_time = random.randint(minimum, maximum)

    self._logger.debug("Sleeping %ds ([%d, %d])" % (sleep_time,
                                                   minimum,
                                                   maximum))

    time.sleep(sleep_time)

  def getCookieValue(self, name, default=None):
    """
    Get the cookie value of the given cookie name.

    @param name: Name of the cookie
    @type name: str
    @param default: Fallback value if the cookie was not found
    @type default: str
    @return: Cookie value
    @rtype: str
    """
    for cookie_name, cookie_value in self.cookies.iteritems():
      if name == cookie_name:
        return cookie_value

    return default

  @property
  def mainForm(self):
    """
    Get the ERP5 main form of the current page. ERP5 generally use
    only one form (whose C{id} is C{main_form}) for all the controls
    within a page. A Form instance is returned including helper
    methods specific to ERP5.

    @return: The main Form class instance
    @rtype: Form

    @raise LookupError: The main form could not be found.

    @todo: Perhaps the page could be parsed to generate a class with
           only appropriate methods, but that would certainly be an
           huge overhead for little benefit...

    @todo: Patch zope.testbrowser to allow the class to be given
           rather than duplicating the code
    """
    # If the page has not changed, no need to re-create a class, so
    # just return the main_form instance
    if self._main_form and self._counter == self._main_form._browser_counter:
      return self._main_form

    main_form = None
    for form in self.mech_browser.forms():
      if form.attrs.get('id') == 'main_form':
        main_form = form

    if not main_form:
      raise LookupError("Could not get 'main_form'")

    self.mech_browser.form = form
    self._main_form = ContextMainForm(self, form)
    return self._main_form

  def getLink(self, text=None, url=None, id=None, index=0,
              class_attribute=None):
    """
    Override original C{getLink} allowing to not consider the HTTP
    query string unless it is explicitly given.

    Also, allows to select a link by its class attribute, which
    basically look for the first element whose C{attribute} is
    C{class_attribute} then call C{getLink} with the element C{href}.

    @param class_attribute: Get the link with this class
    @type class_attribute: str
    @param args: Positional arguments given to original C{getLink}
    @type args: list
    @param kwargs: Keyword arguments given to original C{getLink}
    @type kwargs: dict
    """
    if class_attribute:
      element_list = self.etree.xpath('//a[contains(@class, "%s")]' % \
                                        class_attribute)

      try:
        url = element_list[0].get('href')
      except (IndexError, AttributeError):
        url = None

      if not url:
        raise LookupError("Could not find any link whose class is '%s'" % \
                            class_attribute)

    elif url and '?' not in url:
      url += '?'

    if id is not None:
      def predicate(link):
        return dict(link.attrs).get('id') == id
      args = {'predicate': predicate}
    else:
      import re
      from zope.testbrowser.browser import RegexType

      if isinstance(text, RegexType):
        text_regex = text
      elif text is not None:
        text_regex = re.compile(re.escape(text), re.DOTALL)
      else:
        text_regex = None

      if isinstance(url, RegexType):
        url_regex = url
      elif url is not None:
        url_regex = re.compile(re.escape(url), re.DOTALL)
      else:
        url_regex = None
      args = {'text_regex': text_regex, 'url_regex': url_regex}

    args['nr'] = index
    return LinkWithTime(self.mech_browser.find_link(**args), self)

  def getImportExportLink(self):
    """
    Get Import/Export link. Use the class attribute rather than the
    name as the latter is dependent on the context.

    @return: The link whose class is C{report}
    @rtype: Link

    @todo: Should perhaps be a ContextBrowser class?
    """
    return self.getLink(class_attribute='import_export')

  def getFastInputLink(self):
    """
    Get Fast Input link. Use the class attribute rather than the name
    as the latter is dependent on the context.

    @return: The link whose class is C{fast_input}
    @rtype: Link

    @todo: Should perhaps be a ContextBrowser class?
    """
    return self.getLink(class_attribute='fast_input')

  def getTransitionMessage(self):
    """
    Parse the current page and returns the value of the portal_status
    message.

    @return: The transition message
    @rtype: str

    @raise LookupError: Not found
    """
    try:
      return self.etree.xpath('//div[@id="transition_message"]')[0].text
    except IndexError:
      raise LookupError("Cannot find div with ID 'transition_message'")

  def getInformationArea(self):
    """
    Parse the current page and returns the value of the information_area
    message.

    @return: The information area message
    @rtype: str

    @raise LookupError: Not found
    """
    try:
      return self.etree.xpath('//div[@id="information_area"]')[0].text
    except IndexError:
      raise LookupError("Cannot find div with ID 'information_area'")

  _listbox_table_xpath_str = '//table[contains(@class, "listbox-table")]'

  _legacy_listbox_table_xpath_str = '//div[contains(@class, "listbox")]'\
      '//table'

  def getListboxLink(self, line_number, column_number, cell_element_index=1,
                     *args, **kwargs):
    """
    Follow the link at the given position, excluding any link whose
    class is hidden. In case there are several links within a cell,
    C{cell_element_index} allows to select which one to get (starting
    from 1).

    @param line_number: Line number of the link
    @type line_number: int
    @param column_number: Column number of the link
    @type column_number: int
    @param cell_element_index: Index of the link to be selected in the cell
    @type cell_element_index: int
    @param args: positional arguments given to C{getLink}
    @type args: list
    @param kwargs: keyword arguments given to C{getLink}
    @type kwargs: dict
    @return: C{Link} at the given line and column number
    @rtype: L{zope.testbrowser.interfaces.ILink}

    @raise LookupError: No link could be found at the given position
                        and cell indexes
    """
    if self._is_legacy_listbox:
      listbox_basic_xpath_str = self._legacy_listbox_table_xpath_str
    else:
      listbox_basic_xpath_str = self._listbox_table_xpath_str

    # With XPATH, the position is context-dependent, therefore, as
    # there the cells are either within a <thead> or <tbody>, the line
    # number must be shifted by the number of header lines (namely 2)
    if line_number <= 2:
      relative_line_number = line_number

      if self._is_legacy_listbox:
        column_type = 'td'
      else:
        column_type = 'th'
    else:
      if self._is_legacy_listbox:
        relative_line_number = line_number
      else:
        relative_line_number = line_number - 2

      column_type = 'td'

    xpath_str = '%s//tr[%d]//%s[%d]//a[not(contains(@class, "hidden"))][%d]' % \
        (listbox_basic_xpath_str,
         relative_line_number,
         column_type,
         column_number,
         cell_element_index)

    # xpath() method always return a list even if there is only one element
    element_list = self.etree.xpath(xpath_str)

    try:
      link_href = element_list[0].get('href')
    except (IndexError, AttributeError):
      link_href = None

    if not link_href:
      raise LookupError("Could not find link in listbox cell %dx%d (index=%d)" %\
                          (line_number, column_number, cell_element_index))

    return self.getLink(url=link_href, *args, **kwargs)

  def getListboxPosition(self,
                         text,
                         column_number=None,
                         line_number=None,
                         strict=False):
    """
    Returns the position number of the first line containing given
    text in given column or line number (starting from 1).

    @param text: Text to search
    @type text: str
    @param column_number: Look into all the cells of this column
    @type column_number: int
    @param line_number: Look into all the cells of this line
    @type line_number: int
    @param strict: Should given text matches exactly
    @type strict: bool
    @return: The cell position
    @rtype: int

    @raise LookupError: Not found
    """
    # Require either column_number or line_number to be given
    onlyOne([column_number, line_number], '"column_number" and "line_number"')

    if self._is_legacy_listbox:
      listbox_basic_xpath_str = self._legacy_listbox_table_xpath_str
    else:
      listbox_basic_xpath_str = self._listbox_table_xpath_str

    # Get all cells in the column (if column_number is given and
    # including header columns) or line (if line_number is given)
    if column_number:
      xpath_str_fmt = listbox_basic_xpath_str + '//tr//%%s[%d]' % \
          column_number

      if self._is_legacy_listbox:
        column_or_line_xpath_str = xpath_str_fmt % 'td'
      else:
        column_or_line_xpath_str = "%s | %s" % (xpath_str_fmt % 'th',
                                                xpath_str_fmt % 'td')

    else:
      listbox_basic_xpath_str = self._listbox_table_xpath_str

      # With XPATH, the position is context-dependent, therefore, as
      # there the cells are either within a <thead> or <tbody>, the
      # line number must be shifted by the number of header lines
      # (namely 2)
      if line_number <= 2:
        relative_line_number = line_number

        if self._is_legacy_listbox:
          column_type = 'td'
        else:
          column_type = 'th'
      else:
        if self._is_legacy_listbox:
          relative_line_number = line_number
        else:
          relative_line_number = line_number - 2

        column_type = 'td'

      column_or_line_xpath_str = listbox_basic_xpath_str + '//tr[%d]//%s' %\
          (relative_line_number, column_type)

    cell_list = self.etree.xpath(column_or_line_xpath_str)

    # Iterate over the cells list until one the children content
    # matches the expected text
    for position, cell in enumerate(cell_list):
      for child in cell.iterchildren():
        if not child.text:
          continue

        if (strict and child.text == text) or \
           (not strict and text in child.text):
          return position + 1

    raise LookupError("No matching cell with value '%s'" % text)

  def getRemainingActivityCounter(self):
    """
    Return the number of remaining activities, but do not visit the
    URL so it does not interfere with next calls.

    @return: The number of remaining activities
    @rtype: int
    """
    self._logger.debug("Checking the number of remaining activities")
    activity_counter = self.mech_browser.open_novisit(
      self._erp5_base_url + 'portal_activities/countMessage').read()

    activity_counter = activity_counter and int(activity_counter) or 0
    self._logger.debug("Remaining activities: %d" % activity_counter)

    return activity_counter

  def waitForActivity(self, interval=20, maximum_attempt_number=30):
    """
    Wait for activities every C{interval} seconds at most
    C{maximum_attempt_number} of times and return the waiting time (excluding
    loading time to get the number of remaining time).

    This is mainly relevant when a setup script triggering activities has to
    be executed before running the actual script.

    @param interval: Interval between checking for remaining activities
    @type interval: int
    @param maximum_attempt_number: Number of attempts before failing
    @type maximum_attempt_number: int
    @return: Number of seconds spent waiting
    @rtype: int
    """
    current_attempt_counter = 0
    while current_attempt_counter < maximum_attempt_number:
      if self.getRemainingActivityCounter() == 0:
        return current_attempt_counter * interval

      time.sleep(interval)
      current_attempt_counter += 1

    raise AssertionError("Maximum number of attempts reached while waiting "
                         "for activities to be processed")

from zope.testbrowser.browser import Form, ListControl

class LoginError(Exception):
  """
  Exception raised when login fails
  """
  pass

class MainForm(Form):
  """
  Class defining convenient methods for the main form of ERP5. All the
  methods specified are those always found in an ERP5 page in contrary
  to L{ContextMainForm}.
  """
  __metaclass__ = measurementMetaClass(prefix='submit')

  def submit(self, label=None, name=None, class_attribute=None, index=None,
             *args, **kwargs):
    """
    Overriden for logging purpose, and for specifying a default index
    to 0 if not set, thus avoiding AmbiguityError being raised (in
    ERP5 there may be several submit fields with the same name).

    Also, allows to select a submit by its class attribute, which
    basically look for the first element whose C{attribute} is
    C{class_attribute} then call C{submit} with the element C{name}.

    @param class_attribute: Submit according to the class attribute
    @type class_attribute: str

    @raise LookupError: Could not find any element matching the given
                        class attribute name, if class_attribute
                        parameter is given.
    """
    self._logger.debug(
      "Submitting (name='%s', label='%s', class='%s')" % (name, label,
                                                          class_attribute))

    if class_attribute:
      element_list = self.browser.etree.xpath('//*[contains(@class, "%s")]' % \
                                                class_attribute)

      try:
        name = element_list[0].get('name')
      except (IndexError, AttributeError):
        name = None

      if not name:
        raise LookupError("Could not find any button whose class is '%s'" % \
                            class_attribute)

    if label is None and name is None:
      super(MainForm, self).submit(label=label, name=name, *args, **kwargs)
    else:
      if index is None:
        index = 0

      super(MainForm, self).submit(label=label, name=name, index=index,
                                   *args, **kwargs)

  def submitSelect(self, select_name, submit_name, label=None, value=None,
                   select_index=None, control_index=None):
    """
    Get the select control whose name attribute is C{select_name},
    then select the option control specified either by its C{label} or
    C{value} within that select control, and finally submit it using
    the submit control whose name attribute is C{submit_name}.

    The C{value} matches an option value if found at the end of the
    latter (excluding the query string), for example a search for
    I{/logout} will match I{/erp5/logout} and I{/erp5/logout?foo=bar}
    (if and only if C{value} contains no query string) but not
    I{/erp5/logout_bar}.

    Label value is searched as case-sensitive whole words within the
    labels for each item--that is, a search for I{Add} will match
    I{Add a contact} but not I{Address}.  A word is defined as one or
    more alphanumeric characters or the underline.

    C{select_index} and C{control_index} have the same meaning as in
    zope.testbrowser, namely to select a particular select or control
    when the C{label} or C{value} is ambiguous.

    @param select_name: Select control name
    @type select_name: str
    @param submit_name: Submit control name
    @type submit_name: str
    @param label: Label of the option control
    @type label: str
    @param value: Value of the option control
    @type value: str
    @param select_index: Index of the select if multiple matches
    @type select_index: int
    @param control_index: Index of the control if multiple matches
    @type control_index: int

    @raise LookupError: The select, option or submit control could not
                        be found
    """
    select_control = self.getControl(name=select_name, index=select_index)

    # zope.testbrowser checks for a whole word but it is also useful
    # to match the end of the option control value string because in
    # ERP5, the value could be URL (such as 'http://foo:81/erp5/logout')
    if value:
      for item in select_control.options:
        if '?' not in value:
          item = item.split('?')[0]

        if item.endswith(value):
          value = item
          break

    self._logger.debug("select_id='%s', label='%s', value='%s'" % \
                                 (select_name, label, value))

    select_control.getControl(label=label, value=value,
                              index=control_index).selected = True

    self.submit(name=submit_name)

  def submitLogin(self):
    """
    Log into ERP5 using the username and password provided in the
    browser. It is assumed that the current page is the login page (by
    calling C{open('login_form')} beforehand), otherwise it will open
    that page before logging in.

    Note that if the user is already logged in, it will do nothing.

    @raise LoginError: Login failed

    @todo: Use information sent back as headers rather than looking
           into the page content?
    """
    check_logged_in_xpath = '//div[@id="logged_in_as"]/*'
    if self.etree.xpath(check_logged_in_xpath):
      self._logger.debug("Already logged in")
      # TODO: Perhaps zope.testbrowser should be patched instead?
      self.browser.timer.start_time = self.browser.timer.end_time = 0
      return

    self._logger.debug("Logging in: username='%s', password='%s'" % \
                         (self.browser._username, self.browser._password))

    def login(form):
      form.getControl(name='__ac_name').value = self.browser._username
      form.getControl(name='__ac_password').value = self.browser._password
      form.submit()

    try:
      login(self)
    except LookupError:
      self.browser.open('login_form')
      login(self.browser.mainForm)

    if not self.etree.xpath(check_logged_in_xpath):
      raise LoginError("%s: Could not log in as '%s:%s'" % \
                         (self.browser._erp5_base_url,
                          self.browser._username,
                          self.browser._password))

  def submitSelectFavourite(self, label=None, value=None, **kw):
    """
    Select and submit a favourite, given either by its label (such as
    I{Log out}) or value (I{/logout}). See L{submitSelect}.
    """
    self.submitSelect('select_favorite', 'Base_doFavorite:method', label, value,
                      **kw)

  def submitSelectModule(self, label=None, value=None, **kw):
    """
    Select and submit a module, given either by its label (such as
    I{Currencies}) or value (such as I{/glossary_module}). See
    L{submitSelect}.
    """
    self.submitSelect('select_module', 'Base_doModule:method', label, value,
                      **kw)

  def submitSelectLanguage(self, label=None, value=None, **kw):
    """
    Select and submit a language, given either by its label (such as
    I{English}) or value (such as I{en}). See L{submitSelect}.
    """
    self.submitSelect('select_language', 'Base_doLanguage:method', label, value)

  def submitSearch(self, search_text):
    """
    Fill search field with C{search_text} and submit it.

    @param search_text: Text to search
    @type search_text: str
    """
    self.getControl(name='field_your_search_text').value = search_text
    self.submit(name='ERP5Site_viewQuickSearchResultList:method')

  def submitLogout(self):
    """
    Perform logout.
    """
    self.submitSelectFavourite(value='/logout')

class ContextMainForm(MainForm):
  """
  Class defining context-dependent convenient methods for the main
  form of ERP5.

  @todo:
   - doListboxAction
   - doContextListMode
   - doContextSearch
   - doContextSort
   - doContextConfigure
   - doContextButton
   - doContextReport
   - doContextExchange
  """
  def submitSelectJump(self, label=None, value=None,
                       no_jump_transition_message=None,
                       maximum_attempt_number=1, sleep_between_attempt=0,
                       **kw):
    """
    Select and submit a jump, given either by its label (such as
    I{Queries}) or value (such as
    I{/person_module/Base_jumpToRelatedObject?portal_type=Foo}). See
    L{submitSelect}.

    Usually, a transition message will be displayed if it was not possible to
    jump (for example because the object has not been created yet), therefore
    the number of attempts before failing can be specified if necessary.

    @param no_jump_transition_message: Transition message displayed if the
                                       jump could not be performed
    @type no_jump_transition_message: str
    @param maximum_attempt_number: Number of attempts before failing
    @type maximum_attempt_number: int
    @param sleep_between_attempt: Sleep N seconds between attempts
    @type sleep_between_attempt: int
    """
    if not no_jump_transition_message:
      self.submitSelect('select_jump', 'Base_doJump:method',
                        label, value, **kw)
    else:
      current_attempt_counter = 0
      while current_attempt_counter != maximum_attempt_number:
        self.browser.mainForm.submitSelect('select_jump', 'Base_doJump:method',
                                           label, value, **kw)

        if no_jump_transition_message != self.browser.getTransitionMessage():
          return current_attempt_counter * sleep_between_attempt

        time.sleep(sleep_between_attempt)
        current_attempt_counter += 1

      raise AssertionError("Could not jump to related object")

  def submitSelectAction(self, label=None, value=None, **kw):
    """
    Select and submit an action, given either by its label (such as
    I{Add Person}) or value (such as I{add} and I{add Person}). See
    L{submitSelect}.
    """
    self.submitSelect('select_action', 'Base_doAction:method', label, value,
                      **kw)

  def submitCut(self):
    """
    Cut the previously selected objects.
    """
    self.submit(name='Folder_cut:method')

  def submitCopy(self):
    """
    Copy the previously selected objects.
    """
    self.submit(name='Folder_copy:method')

  def submitPaste(self):
    """
    Paste the previously selected objects.
    """
    self.submit(name='Folder_paste:method')

  def submitPrint(self):
    """
    Print the previously selected objects. Use the class attribute
    rather than the name as the latter is dependent on the context.
    """
    self.submit(class_attribute='print')

  def submitReport(self):
    """
    Create a report. Use the class attribute rather than the name as
    the latter is dependent on the context.
    """
    self.submit(class_attribute='report')

  def submitNew(self):
    """
    Create a new object.
    """
    self.submit(name='Folder_create:method')

  def submitClone(self):
    """
    Clone the previously selected objects. Use the class attribute
    rather than the name as the latter is dependent on the context.
    """
    self.submit(class_attribute='clone')

  def submitDelete(self):
    """
    Delete the previously selected objects.
    """
    self.submit(name='Folder_deleteObjectList:method')

  def submitSave(self):
    """
    Save the previously selected objects.
    """
    self.submit(name='Base_edit:method')

  def submitShow(self):
    """
    Show the previously selected objects.
    """
    self.submit(name='Folder_show:method')

  def submitFilter(self):
    """
    Filter the objects.
    """
    self.submit(name='Folder_filter:method')

  def submitAction(self):
    """
    Select/unselect objects.
    """
    self.submit(name='Base_doSelect:method')

  def submitSelectWorkflow(self, label=None, value=None,
                           script_id='viewWorkflowActionDialog',
                           maximum_attempt_number=1,
                           sleep_between_attempt=0,
                           dialog_name=None,
                           dialog_expected_transition_message=None,
                           **kw):
    """
    Select and submit a workflow action, given either by its label
    (such as I{Create User}) or value (such as I{create_user_action}
    in I{/Person_viewCreateUserActionDialog?workflow_action=create_user_action},
    with C{script_id=Person_viewCreateUserActionDialog}). See L{submitSelect}.

    When validating an object, L{submitDialogConfirm} allows to perform the
    validation required on the next page, and can be called directly by
    passing the function object in C{dialog_name} parameter.

    As the Workflow action may not be available yet, it is possible to set the
    maximum number of attempts and the sleep duration between each attempt.

    The return value depends on the parameters (see C{timeInSecondDecorator}):
    - C{dialog_name} is given:
      - If C{maximum_attempt_number} equals to 1:
        (show_page_after_dialog_time, show_dialog_time)
      - Otherwise:
        (show_page_after_dialog_time, show_dialog_time, wait_time)
    - Otherwise:
      - If C{maximum_attempt_number} equals to 1:
        show_dialog_time
      - Otherwise:
        (show_dialog_time, wait_time)

    @param script_id: Script identifier
    @type script_id: str
    @param maximum_attempt_number: Number of attempts before failing
    @type maximum_attempt_number: int
    @param sleep_between_attempt: Sleep N seconds between attempts
    @type sleep_between_attempt: int
    @param dialog_name: Function to call after the workflow action ('cancel' or 'confirm')
    @type dialog_name: str
    @param dialog_expected_transition_message: Expected dialog transition message
    @type dialog_expected_transition_message: str
    """
    url_before = self.browser.url

    def tryLegacyAndNew():
      try:
        self.browser.mainForm.submitSelect(
          'select_action', 'Base_doAction:method', label,
          value and '%s?workflow_action=%s' % (script_id, value), **kw)

      except LookupError:
        self.browser.mainForm.submitSelect(
          'select_action', 'Base_doAction:method', label,
          value and '%s?field_my_workflow_action=%s' % (script_id, value), **kw)

      if dialog_name:
        show_dialog_time = self.lastRequestSeconds

        # TODO: Should it fail completely if the dialog could not be found?
        getattr(self.browser.mainForm,
                'submitDialog' + dialog_name.capitalize())()

        if dialog_expected_transition_message:
          transition_message = self.browser.getTransitionMessage()
          if transition_message != dialog_expected_transition_message:
            raise AssertionError("Expected transition message: %s, got: %s" % \
                                   (dialog_expected_transition_message,
                                    transition_message))

        return show_dialog_time

    if maximum_attempt_number == 1:
      return tryLegacyAndNew()
    else:
      show_dialog_time = None
      current_attempt_number = 1
      while True:
        try:
          show_dialog_time = tryLegacyAndNew()
        except (AssertionError, LookupError):
          if current_attempt_number == maximum_attempt_number:
            raise

          current_attempt_number += 1
          time.sleep(sleep_between_attempt)

          # The page needs to be reloaded before the next attempt
          self.browser.open(url_before)
        else:
          break

      wait_time = (current_attempt_number - 1) * sleep_between_attempt
      if show_dialog_time is not None:
        return (show_dialog_time, wait_time)

      return wait_time

  def submitDialogCancel(self):
    """
    Cancel the dialog action. A dialog is showed when validating a
    workflow or deleting an object for example.
    """
    self.submit(name='Base_cancel:method')

  def submitDialogUpdate(self):
    """
    Update the dialog action. A dialog may contain a button to update
    the form before confirming it. See L{submitDialogConfirm} as well.
    """
    self.submit(name='Base_showUpdateDialog:method')

  def submitDialogConfirm(self):
    """
    Confirm the dialog action. A dialog is showed when validating a
    workflow or deleting an object for example.

    @todo: Specifying index is kind of ugly (there is C{dummy} field
           with the same name though)
    """
    self.submit(name='Base_callDialogMethod:method')

  def getListboxControl(self, line_number, column_number, cell_element_index=1,
                        *args, **kwargs):
    """
    Get the control located at line and column numbers (both starting
    from 1), excluding hidden control and those whose class is hidden
    too. The position of a cell from a column or line number can be
    obtained through calling
    L{erp5.util.testbrowser.browser.Browser.getListboxPosition}.

    Also, there may be several elements within a cell, thus
    C{cell_element_index} allows to select which one to get (starting
    from 1).

    @param line_number: Line number of the field
    @type line_number: int
    @param column_number: Column number of the field
    @type column_number: int
    @param cell_element_index: Index of the control to be selected in the cell
    @type cell_element_index: int
    @param args: positional arguments given to the parent C{getControl}
    @type args: list
    @param kwargs: keyword arguments given to the parent C{getControl}
    @type kwargs: dict
    @return: The control found at the given line and column numbers
    @rtype: L{zope.testbrowser.interfaces.IControl}

    @raise LookupError: No control could be found at the given
                        position and cell indexes
    """
    if self.browser._is_legacy_listbox:
      listbox_basic_xpath_str = self.browser._legacy_listbox_table_xpath_str
    else:
      listbox_basic_xpath_str = self.browser._listbox_table_xpath_str

    if line_number <= 2:
      relative_line_number = line_number

      if self.browser._is_legacy_listbox:
        column_type = 'td'
      else:
        column_type = 'th'
    else:
      if self.browser._is_legacy_listbox:
        relative_line_number = line_number
      else:
        relative_line_number = line_number - 2

      column_type = 'td'

    xpath_str = '%s//tr[%d]//%s[%d]/*[not(@type="hidden") and ' \
        'not(contains(@class, "hidden"))][%d]' % \
        (listbox_basic_xpath_str,
         relative_line_number,
         column_type,
         column_number,
         cell_element_index)

    # xpath() method always return a list even if there is only one element
    element_list = self.browser.etree.xpath(xpath_str)

    try:
      input_element = element_list[0]
      input_name = input_element.get('name')
    except (IndexError, AttributeError):
      input_element = input_name = None

    if input_element is None or not input_name:
      raise LookupError("Could not find control in listbox cell %dx%d (index=%d)" %\
                          (line_number, column_number, cell_element_index))

    control = self.getControl(name=input_element.get('name'), *args, **kwargs)

    # If this is a list control (radio button, checkbox or select
    # control), then get the item from its value
    if isinstance(control, ListControl):
      control = control.getControl(value=input_element.get('value'))

    return control

from zope.testbrowser.browser import SubmitControl

class SubmitControlWithTime(SubmitControl):
  """
  Only define to wrap click methods to measure the time spent
  """
  __metaclass__ = measurementMetaClass(prefix='click')

from zope.testbrowser.browser import ImageControl

class ImageControlWithTime(ImageControl):
  """
  Only define to wrap click methods to measure the time spent
  """
  __metaclass__ = measurementMetaClass(prefix='click')

import zope.testbrowser.browser

browser_controlFactory = zope.testbrowser.browser.controlFactory
def controlFactory(control, *args, **kwargs):
  """
  Monkey patch controlFactory in zope.testbrowser to get elapsed time on
  ImageControl and SubmitControl
  """
  try:
    t = control.type
  except AttributeError:
    # This is a subcontrol
    pass
  else:
    if t in ('submit', 'submitbutton'):
      return SubmitControlWithTime(control, *args, **kwargs)
    elif t == 'image':
      return ImageControlWithTime(control, *args, **kwargs)

  return browser_controlFactory(control, *args, **kwargs)

zope.testbrowser.browser.controlFactory = controlFactory

from zope.testbrowser.browser import Link

class LinkWithTime(Link):
  """
  Only define to wrap click methods to measure the time spent
  """
  __metaclass__ = measurementMetaClass(prefix='click')



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#                    Arnaud Fontaine <arnaud.fontaine@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import sys
import math
import os
import csv
import logging
import signal
import traceback

class BenchmarkResultStatistic(object):
  def __init__(self, suite, label):
    self.suite = suite
    self.label = label

    self.full_label = '%s: %s' % (self.suite, self.label)

    self.minimum = sys.maxint
    self.maximum = -1
    self.n = 0
    self.error_sum = 0

    # For calculating the mean
    self._value_sum = 0

    # For calculating the standard deviation
    self._variance_sum = 0
    self._variance_mean = 0

  def __str__(self):
    return "%s: min=%.3f, mean=%.3f (+/- %.3f), max=%.3f" % \
        (self.full_label,
         self.minimum,
         self.mean,
         self.standard_deviation,
         self.maximum)

  def add(self, value):
    if value < 0:
      self.error_sum += 1
      return

    if value < self.minimum:
      self.minimum = value
    if value > self.maximum:
      self.maximum = value

    self._value_sum += value
    self.n += 1

    delta = value - self._variance_mean
    self._variance_mean += delta / self.n
    self._variance_sum += delta * (value - self._variance_mean)

  @property
  def mean(self):
    return self._value_sum / self.n

  @property
  def standard_deviation(self):
    return math.sqrt(self._variance_sum / self.n)

class NothingFlushedException(Exception):
  pass

import abc
import time

class BenchmarkResult(object):
  __metaclass__ = abc.ABCMeta

  def __init__(self, argument_namespace, nb_users, user_index,
               current_repeat_range):
    self._argument_namespace = argument_namespace
    self._nb_users = nb_users
    self._user_index = user_index
    self._current_repeat_range = current_repeat_range

    self._logger = None
    self._label_list = None

    self._all_suite_list = []
    self._current_suite_index = 0
    self._all_result_list = []
    self._current_suite_dict = None
    self._current_result_list = None

  @property
  def logger(self):
    if not self._logger:
      logging.basicConfig(
        stream=self.log_file,
        level=(self._argument_namespace.enable_debug and
               logging.DEBUG or logging.INFO),
        format='%(asctime)s: %(levelname)s: %(message)s [%(name)s]')

      self._logger = logging.getLogger('erp5.util.benchmark')
      return self._logger

    return self._logger

  def __enter__(self):
    return self

  def enterSuite(self, name):
    self._current_use_case_elapsed_time = time.time()
    self._current_use_case_counter = 0

    try:
      self._current_suite_dict = self._all_suite_list[self._current_suite_index]

    except IndexError:
      self._current_suite_dict = {
        'name': name,
        'all_result_list': [],
        'stat_list': [],
        # Number of expected results
        'expected': -1,
        'all_use_case_result_list': [],
        'use_case_stat': BenchmarkResultStatistic(name, 'Use cases/minutes')}

      self._all_suite_list.append(self._current_suite_dict)

    self._current_result_list = []

  def __call__(self, label, value):
    try:
      result_statistic = \
          self._current_suite_dict['stat_list'][len(self._current_result_list)]
    except IndexError:
      result_statistic = BenchmarkResultStatistic(
        self._current_suite_dict['name'], label)

      self._current_suite_dict['stat_list'].append(result_statistic)

    result_statistic.add(value)
    self._current_result_list.append(value)

  def incrementCurrentSuiteUseCase(self):
    self._current_use_case_counter += 1

  @property
  def label_list(self):
    if self._label_list:
      return self._label_list

    # TODO: Should perhaps be cached...
    label_list = []
    for suite_dict in self._all_suite_list:
      if suite_dict['expected'] == -1:
        return None

      suite_label_list = [ stat.full_label for stat in suite_dict['stat_list']]
      suite_label_list.extend(('Use cases', 'Use cases seconds elapsed'))

      label_list.extend(suite_label_list)

    self._label_list = label_list
    return label_list

  def getCurrentSuiteStatList(self):
    return self._current_suite_dict['stat_list']

  def getCurrentSuiteUseCaseStat(self):
    return self._current_suite_dict['use_case_stat']

  @staticmethod
  def _addResultWithError(result_list, expected_len):
    missing_result_n = expected_len - len(result_list)
    if missing_result_n > 0:
      result_list.extend(missing_result_n * [-1])

  def exitSuite(self, with_error=False):
    elapsed_time = time.time() - self._current_use_case_elapsed_time

    if with_error:
      if self._current_suite_dict['expected'] != -1:
        self._addResultWithError(self._current_result_list,
                                 self._current_suite_dict['expected'])

    else:
      if self._current_use_case_counter == 0:
        self._current_use_case_counter = 1

      if self._current_suite_dict['expected'] == -1:
        self._current_suite_dict['expected'] = len(self._current_result_list)

        # Fix previous results
        for result_list in self._current_suite_dict['all_result_list']:
          self._addResultWithError(result_list,
                                   self._current_suite_dict['expected'])

    self._current_suite_dict['all_use_case_result_list'].append(
      (self._current_use_case_counter, int(elapsed_time)))

    if self._current_use_case_counter != 0:
      self._current_suite_dict['use_case_stat'].add(
        self._current_use_case_counter / (elapsed_time / 60.0))

    self._current_suite_dict['all_result_list'].append(self._current_result_list)
    self._current_suite_index += 1

  def iterationFinished(self):
    self._current_suite_index = 0

  @abc.abstractmethod
  def flush(self, partial=True):
    # TODO: Should perhaps be cached...
    all_result_list = []
    for result_dict in self._all_suite_list:
      if result_dict['expected'] == -1:
        raise NothingFlushedException()

      for index, result_list in enumerate(result_dict['all_result_list']):
        result_list.extend(result_dict['all_use_case_result_list'][index])

        try:
          all_result_list[index].extend(result_list)
        except IndexError:
          all_result_list.append(result_list)

      result_dict['all_result_list'] = []
      result_dict['all_use_case_result_list'] = []

    return all_result_list

  @abc.abstractmethod
  def __exit__(self, exc_type, exc_value, traceback_object):
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

    try:
      self.flush(partial=False)
    except NothingFlushedException:
      pass

    return True

class CSVBenchmarkResult(BenchmarkResult):
  def __init__(self, *args, **kwargs):
    super(CSVBenchmarkResult, self).__init__(*args, **kwargs)

    filename_prefix = self._getFilenamePrefix()

    self._result_filename = "%s.csv" % filename_prefix
    self._result_filename_path = os.path.join(
      self._argument_namespace.report_directory, self._result_filename)

    self._log_filename = "%s.log" % filename_prefix
    self._log_filename_path = os.path.join(
      self._argument_namespace.report_directory, self._log_filename)

    self.log_file = open(self._log_filename_path, 'w')

  def _getFilenamePrefix(self):
    if isinstance(self._argument_namespace.users, int):
      max_nb_users = self._argument_namespace.users
      suffix = ''
    else:
      max_nb_users = self._argument_namespace.users[1]
      suffix = '-rrepeat%%0%dd' % len(str(self._argument_namespace.repeat_range))

      suffix = suffix % self._current_repeat_range

    fmt = "%%s-%%drepeat-%%0%ddusers-process%%0%dd%s" % \
        (len(str(max_nb_users)), len(str(self._nb_users)),
         suffix)

    return fmt % (self._argument_namespace.filename_prefix,
                  self._argument_namespace.repeat,
                  self._nb_users,
                  self._user_index)

  def __enter__(self):
    self._result_file = open(self._result_filename_path, 'wb')
    self._csv_writer = csv.writer(self._result_file, delimiter=',',
                                  quoting=csv.QUOTE_MINIMAL)

    return self

  def flush(self, partial=True):
    result_list = super(CSVBenchmarkResult, self).flush(partial)

    if self._result_file.tell() == 0:
      self._csv_writer.writerow(self.label_list)

    self._csv_writer.writerows(result_list)
    self._result_file.flush()
    os.fsync(self._result_file.fileno())

  def __exit__(self, exc_type, exc_value, traceback_object):
    super(CSVBenchmarkResult, self).__exit__(exc_type, exc_value,
                                             traceback_object)
    self._result_file.close()

    if exc_type and not issubclass(exc_type, StopIteration):
      msg = "An error occured, see: %s" % self._log_filename_path

      logged_msg = str(exc_value)
      if not issubclass(exc_type, RuntimeError):
        logged_msg += ''.join(traceback.format_tb(traceback_object))

      self.logger.error(logged_msg)
      raise RuntimeError(msg)

from cStringIO import StringIO

import xmlrpclib
import datetime

class ERP5BenchmarkResult(BenchmarkResult):
  def __init__(self, *args, **kwargs):
    self.log_file = StringIO()
    self._log_buffer_list = []

    super(ERP5BenchmarkResult, self).__init__(*args, **kwargs)

  def iterationFinished(self):
    super(ERP5BenchmarkResult, self).iterationFinished()

    # TODO: garbage?
    self._log_buffer_list.append(self.log_file.getvalue())
    self.log_file.seek(0)

  def flush(self, partial=True):
    result_list = super(ERP5BenchmarkResult, self).flush()

    benchmark_result = xmlrpclib.ServerProxy(
      self._argument_namespace.erp5_publish_url,
      verbose=True,
      allow_none=True)

    benchmark_result.BenchmarkResult_addResultLineList(
      self._argument_namespace.user_tuple[self._user_index][0],
      self._argument_namespace.repeat,
      self._nb_users,
      self._argument_namespace.benchmark_suite_name_list,
      self.label_list,
      result_list,
      self._log_buffer_list)

  def __exit__(self, exc_type, exc_value, traceback_object):
    super(ERP5BenchmarkResult, self).__exit__(exc_type, exc_value,
                                              traceback_object)

  @staticmethod
  def createResultDocument(publish_url, publish_project, repeat, nb_users):
    test_result_module = xmlrpclib.ServerProxy(publish_url,
                                               verbose=True,
                                               allow_none=True)

    if isinstance(nb_users, tuple):
      nb_users_str = '%d to %d' % nb_users
    else:
      nb_users_str = '%d' % nb_users

    benchmark_result = test_result_module.TestResultModule_addBenchmarkResult(
      '%d repeat with %s concurrent users' % (repeat, nb_users_str),
      publish_project, ' '.join(sys.argv), datetime.datetime.now())

    return benchmark_result['id']

  @staticmethod
  def closeResultDocument(publish_document_url, error_message_set):
    result = xmlrpclib.ServerProxy(publish_document_url,
                                   verbose=True,
                                   allow_none=True)

    result.BenchmarkResult_completed(error_message_set and 'FAIL' or 'PASS',
                                     error_message_set)
